import unittest

from pymesomb import mesomb
from pymesomb.exceptions import ServiceNotFoundException, PermissionDeniedException, InvalidClientRequestException
from pymesomb.operations import PaymentOperation


class YangoTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.application_key = '4a25f8788665924aa2905daf3b2700eb916c7d56'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_make_payment_with_not_found_service(self):
        operation = PaymentOperation(self.application_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.make_yango_refill(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf', driver_id='8cadabd4c759446a9f9e4042d628ee1f')

    def test_make_payment_with_permission_denied(self):
        operation = PaymentOperation(self.application_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.make_yango_refill(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf',
                                        driver_id='8cadabd4c759446a9f9e4042d628ee1f')

    def test_make_payment_with_invalid_amount(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        with self.assertRaises(InvalidClientRequestException):
            operation.make_yango_refill(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf',
                                        driver_id='8cadabd4c759446a9f9e4042d628ee1f')

    def test_make_payment_success(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_yango_refill(amount=100, service='MTN', payer='670000000', trx_id='1', driver_id='8cadabd4c759446a9f9e4042d628ee1f')
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_transaction_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.transaction.amount, 98.5)
        self.assertEqual(response.transaction.fees, 1.5)
        self.assertEqual(response.transaction.b_party, "237670000000")
        self.assertEqual(response.transaction.country, "CM")
        self.assertEqual(response.transaction.currency, "XAF")
        self.assertEqual(response.transaction.reference, "1")

    def test_make_payment_success_and_products(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_yango_refill(amount=100, service='MTN', payer='670000000', driver_id='8cadabd4c759446a9f9e4042d628ee1f', trx_id='1', customer={
            'phone': '+237677550439',
            'email': 'fisher.bank@gmail.com',
            'first_name': 'Fisher',
            'last_name': 'BANK',
        }, location={
            'town': 'Douala',
            'region': 'Littoral',
            'country': 'Cameroun'
        })
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_transaction_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.transaction.amount, 98.5)
        self.assertEqual(response.transaction.fees, 1.5)
        self.assertEqual(response.transaction.b_party, "237670000000")
        self.assertEqual(response.transaction.country, "CM")
        self.assertEqual(response.transaction.currency, "XAF")
        self.assertEqual(response.transaction.reference, "1")
        self.assertEqual(response.transaction.customer.phone, "+237677550439")
        self.assertEqual(response.transaction.customer.email, "fisher.bank@gmail.com")
        self.assertEqual(response.transaction.customer.first_name, "Fisher")
        self.assertEqual(response.transaction.customer.last_name, "BANK")
        self.assertEqual(response.transaction.location.town, "Douala")
        self.assertEqual(response.transaction.location.country, "Cameroun")

    def test_make_payment_pending(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_yango_refill(amount=100, service='MTN', payer='670000000', mode='asynchronous', driver_id='8cadabd4c759446a9f9e4042d628ee1f')
        self.assertTrue(response.is_operation_success())
        self.assertFalse(response.is_transaction_success())
        self.assertEqual(response.transaction.status, "PENDING")


import unittest

from pymesomb import mesomb
from pymesomb.exceptions import ServiceNotFoundException, PermissionDeniedException, InvalidClientRequestException
from pymesomb.operations import FundraisingOperation


class Contribution(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.fund_key = 'fa78bded201b791712ee398c7ddfb8652669404f'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_make_payment_with_not_found_service(self):
        operation = FundraisingOperation(self.fund_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.make_contribution(amount=5, service='MTN', payer='670000000',
                                        full_name={'first_name': 'John', 'last_name': 'Doe'},
                                        contact={'email': 'contact@gmail.com', 'phone_number': '+237677550203'})

    def test_make_payment_with_permission_denied(self):
        operation = FundraisingOperation(self.fund_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.make_contribution(amount=5, service='MTN', payer='670000000',
                                        full_name={'first_name': 'John', 'last_name': 'Doe'},
                                        contact={'email': 'contact@gmail.com', 'phone_number': '+237677550203'})

    def test_make_payment_with_invalid_amount(self):
        operation = FundraisingOperation(self.fund_key, self.access_key, self.secret_key)
        with self.assertRaises(InvalidClientRequestException):
            operation.make_contribution(amount=5, service='MTN', payer='670000000',
                                        full_name={'first_name': 'John', 'last_name': 'Doe'},
                                        contact={'email': 'contact@gmail.com', 'phone_number': '+237677550203'})

    def test_make_contribution_success(self):
        operation = FundraisingOperation(self.fund_key, self.access_key, self.secret_key)
        response = operation.make_contribution(amount=100, service='MTN', payer='670000000',
                                               full_name={'first_name': 'John', 'last_name': 'Doe'},
                                               contact={'email': 'contact@gmail.com', 'phone_number': '+237677550203'})
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_contribution_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.contribution.amount, 98)
        self.assertEqual(response.contribution.fees, 2)
        self.assertEqual(response.contribution.b_party, "237670000000")
        self.assertEqual(response.contribution.country, "CM")
        self.assertEqual(response.contribution.currency, "XAF")
        self.assertEqual(response.contribution.contributor.phone, "+237677550203")
        self.assertEqual(response.contribution.contributor.email, "contact@gmail.com")
        self.assertEqual(response.contribution.contributor.first_name, "John")
        self.assertEqual(response.contribution.contributor.last_name, "Doe")

    def test_make_contribution_anonymously_success(self):
        operation = FundraisingOperation(self.fund_key, self.access_key, self.secret_key)
        response = operation.make_contribution(amount=100, service='MTN', payer='670000000', anonymous=True)

        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_contribution_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.contribution.amount, 98)
        self.assertEqual(response.contribution.fees, 2)
        self.assertEqual(response.contribution.b_party, "237670000000")
        self.assertEqual(response.contribution.country, "CM")
        self.assertEqual(response.contribution.currency, "XAF")
        self.assertIsNone(response.contribution.contributor)


class ContributionTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.fund_key = 'fa78bded201b791712ee398c7ddfb8652669404f'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_get_contributions_with_no_found_service(self):
        operation = FundraisingOperation(self.fund_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.get_contributions(['c6c40b76-8119-4e93-81bf-bfb55417b392'])

    def test_get_contributions_with_permission_denied(self):
        operation = FundraisingOperation(self.fund_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.get_contributions(['c6c40b76-8119-4e93-81bf-bfb55417b392'])

    def test_get_contributions_success(self):
        operation = FundraisingOperation(self.fund_key, self.access_key, self.secret_key)
        contributions = operation.get_contributions(['0685831f-4145-4352-ae81-155fec42c748'])
        self.assertEqual(1, len(contributions))

    def test_check_contributions_success(self):
        operation = FundraisingOperation(self.fund_key, self.access_key, self.secret_key)
        contributions = operation.check_contributions(['0685831f-4145-4352-ae81-155fec42c748'])
        self.assertEqual(1, len(contributions))



import random
import re
import string


class RandomGenerator:
    """ """
    @staticmethod
    def nonce(length=40):
        """

        Args:
          length:  (Default value = 40)

        Returns:

        """
        letters = string.ascii_letters + string.digits
        return ''.join(random.choice(letters) for i in range(length))


def detect_operator(phone, country='CM'):
    """Detect the operator of a phone number in a country (only cameroon is supported for now)

    Args:
      phone: 
      country:  (Default value = 'CM')

    Returns:

    """
    OPERATOR_REGEX = {
        'MTN': r'^(237)?(67|65[0-4]|68[0-3])',
        'ORANGE': r'^(237)?(69|65[5-9])',
        'NEXTTEL': r'^(237)?(66)',
        'YOOMEE': r'^(237)?(242)',
        'CAMTEL': r'^(237)?(233|222|243|62)',
        'MESOMB': r'^7',
    }
    for operator, regex in OPERATOR_REGEX.items():
        if re.match(regex, phone):
            return operator



import hashlib
import hmac
import json
from urllib.parse import quote, urlparse, quote_plus

from pymesomb import mesomb


class Signature:
    """ """
    @staticmethod
    def sign_request(service, method, url, date, nonce, credentials, headers=None, body=None):
        """Method to use to compute signature used in MeSomb request

        Args:
          service: service to use can be payment, wallet ... (the list is provide by MeSomb)
          method: HTTP method (GET, POST, PUT, PATCH, DELETE...)
          url: the full url of the request with query element
          date: Datetime of the request
          nonce: Unique string generated for each request sent to MeSomb
          credentials: dict containing key => value for the credential provided by MeSOmb.
                {'access' => access_key, 'secret' => secret_key}
          headers: Extra HTTP header to use in the signature (Default value = None)
          body: The dict containing the body you send in your request body (Default value = None)

        Returns:
          Authorization to put in the header

        """
        algorithm = mesomb.algorithm
        parse = urlparse(url)
        canonical_query = parse.query
        if len(canonical_query) > 0:
            canonical_query = '&'.join([f"{quote_plus(c.split('=')[0])}={quote_plus(c.split('=')[1])}" for c in canonical_query.split('&')])

        timestamp = str(int(date.timestamp()))

        # CanonicalHeaders
        if headers is None:
            headers = {}
        headers['host'] = f'{parse.scheme}://{parse.netloc}'
        headers['x-mesomb-date'] = timestamp
        headers['x-mesomb-nonce'] = nonce
        canonical_headers = '\n'.join([f'{key.lower()}:{headers[key].strip()}' for key in sorted(headers)])

        if body is None:
            body = {}
        request_params = json.dumps(body, separators=(',', ':'))

        payload_hash = hashlib.sha1(request_params.encode('utf-8')).hexdigest()

        signed_headers = ';'.join(sorted(headers))

        canonical_request = f'{method}\n{quote(parse.path)}\n{canonical_query}\n{canonical_headers}\n{signed_headers}\n{payload_hash}'

        scope = f"{date.strftime('%Y%m%d')}/{service}/mesomb_request"
        string_to_sign = f"{algorithm}\n{timestamp}\n{scope}\n{hashlib.sha1(canonical_request.encode('utf-8')).hexdigest()}"

        signature = hmac.new(credentials['secret_key'].encode(), string_to_sign.encode(), hashlib.sha1).hexdigest()

        authorization_header = f"{algorithm} Credential={credentials['access_key']}/{scope}, SignedHeaders={signed_headers}, Signature={signature}"

        return authorization_header



import unittest


class UtilTest(unittest.TestCase):
    def test_operator_detect(self):
        from pymesomb.utils import detect_operator

        self.assertEqual(detect_operator('677559230'), 'MTN')
        self.assertEqual(detect_operator('237677559230'), 'MTN')
        self.assertEqual(detect_operator('690090980'), 'ORANGE')
        self.assertEqual(detect_operator('237690090980'), 'ORANGE')


import unittest

from pymesomb import mesomb
from pymesomb.exceptions import ServiceNotFoundException, PermissionDeniedException, InvalidClientRequestException
from pymesomb.operations import PaymentOperation


class CollectTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.application_key = '2bb525516ff374bb52545bf22ae4da7d655ba9fd'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_make_payment_with_not_found_service(self):
        operation = PaymentOperation(self.application_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.make_collect(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf')

    def test_make_payment_with_permission_denied(self):
        operation = PaymentOperation(self.application_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.make_collect(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf')

    def test_make_payment_with_invalid_amount(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        with self.assertRaises(InvalidClientRequestException):
            operation.make_collect(amount=5, service='MTN', payer='670000000', nonce='ifshdifsdf')

    def test_make_payment_success(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_collect(amount=100, service='MTN', payer='670000000', trx_id='1')
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_transaction_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.transaction.amount, 98)
        self.assertEqual(response.transaction.fees, 2)
        self.assertEqual(response.transaction.b_party, "237670000000")
        self.assertEqual(response.transaction.country, "CM")
        self.assertEqual(response.transaction.currency, "XAF")
        self.assertEqual(response.transaction.reference, "1")

    def test_make_payment_success_and_products(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_collect(amount=100, service='MTN', payer='670000000', trx_id='1', customer={
            'phone': '+237677550439',
            'email': 'fisher.bank@gmail.com',
            'first_name': 'Fisher',
            'last_name': 'BANK',
        }, location={
            'town': 'Douala',
            'region': 'Littoral',
            'country': 'Cameroun'
        }, products=[
            {
                'id': 'SKU001',
                'name': 'Sac a Main',
                'category': 'Sac',
            }
        ])
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_transaction_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.transaction.amount, 98.0)
        self.assertEqual(response.transaction.fees, 2)
        self.assertEqual(response.transaction.b_party, "237670000000")
        self.assertEqual(response.transaction.country, "CM")
        self.assertEqual(response.transaction.currency, "XAF")
        self.assertEqual(response.transaction.reference, "1")
        self.assertEqual(response.transaction.customer.phone, "+237677550439")
        self.assertEqual(response.transaction.customer.email, "fisher.bank@gmail.com")
        self.assertEqual(response.transaction.customer.first_name, "Fisher")
        self.assertEqual(response.transaction.customer.last_name, "BANK")
        self.assertEqual(response.transaction.location.town, "Douala")
        self.assertEqual(response.transaction.location.country, "Cameroun")
        self.assertEqual(len(response.transaction.products), 1)

    def test_make_payment_pending(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_collect(amount=100, service='MTN', payer='670000000', mode='asynchronous')
        self.assertTrue(response.is_operation_success())
        self.assertFalse(response.is_transaction_success())
        self.assertEqual(response.transaction.status, "PENDING")


class DepositTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://192.168.100.10:8000'
        self.application_key = '2bb525516ff374bb52545bf22ae4da7d655ba9fd'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_make_deposit_with_not_found_service(self):
        operation = PaymentOperation(self.application_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.make_deposit(amount=5, service='MTN', receiver='670000000', nonce='ifshdifsdf')

    def test_make_deposit_with_permission_denied(self):
        operation = PaymentOperation(self.application_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.make_deposit(amount=5, service='MTN', receiver='670000000', nonce='ifshdifsdf')

    def test_make_deposit_with_invalid_amount(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        with self.assertRaises(InvalidClientRequestException):
            operation.make_deposit(amount=5, service='MTN', receiver='670000000', nonce='ifshdifsdf')

    def test_make_deposit_success(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        response = operation.make_deposit(amount=100, service='MTN', receiver='670000000', trx_id='1')
        self.assertTrue(response.is_operation_success())
        self.assertTrue(response.is_transaction_success())
        self.assertEqual(response.status, "SUCCESS")
        self.assertEqual(response.transaction.amount, 100)
        self.assertEqual(response.transaction.fees, 0)
        self.assertEqual(response.transaction.b_party, "237670000000")
        self.assertEqual(response.transaction.country, "CM")
        self.assertEqual(response.transaction.currency, "XAF")
        self.assertEqual(response.transaction.reference, "1")


class StatusTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.application_key = '2bb525516ff374bb52545bf22ae4da7d655ba9fd'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_get_status_with_no_found_service(self):
        operation = PaymentOperation(self.application_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.get_status()

    def test_get_status_with_permission_denied(self):
        operation = PaymentOperation(self.application_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.get_status()

    def test_get_status_success(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        application = operation.get_status()
        self.assertEqual(application.name, 'Meudocta Shop')
        self.assertListEqual(application.countries, ['CM', 'NE'])


class TransactionTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.application_key = '2bb525516ff374bb52545bf22ae4da7d655ba9fd'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key = 'fe8c2445-810f-4caa-95c9-778d51580163'

    def test_get_transactions_with_no_found_service(self):
        operation = PaymentOperation(self.application_key + 'f', self.access_key, self.secret_key)
        with self.assertRaises(ServiceNotFoundException):
            operation.get_transactions(['c6c40b76-8119-4e93-81bf-bfb55417b392'])

    def test_get_transactions_with_permission_denied(self):
        operation = PaymentOperation(self.application_key, 'f' + self.access_key, self.secret_key)
        with self.assertRaises(PermissionDeniedException):
            operation.get_transactions(['c6c40b76-8119-4e93-81bf-bfb55417b392'])

    def test_get_transactions_success(self):
        operation = PaymentOperation(self.application_key, self.access_key, self.secret_key)
        transactions = operation.get_transactions(['9886f099-dee2-4eaa-9039-e92b2ee33353', '9886f099-dee2-4eaa-9039-e92b2ee33354'])
        self.assertEqual(1, len(transactions))


import json
from abc import ABC
from datetime import datetime
from typing import Optional, Dict, List, Any

import requests

from pymesomb import mesomb, __version__
from pymesomb.models import (TransactionResponse, Application, Transaction, Wallet, PaginatedWallets,
                             WalletTransaction, PaginatedWalletTransactions, ContributionResponse, Contribution)
from pymesomb.signature import Signature
from pymesomb.utils import RandomGenerator


class AOperation(ABC):
    """ """
    service = None

    def __init__(self, target, access_key, secret_key, language='en'):
        self.target = target
        self.access_key = access_key
        self.secret_key = secret_key
        self.language = language

    def process_client_exception(self, response):
        """
        Process exception from the client request

        Args:
            response: the response from the request

        Raises:
            ServiceNotFoundException: When the service is not found
            PermissionDeniedException: When the permission is denied
            InvalidClientRequestException: When the client request is invalid
            ServerException: When the server return an error
        """
        from pymesomb.exceptions import ServiceNotFoundException, PermissionDeniedException, \
            InvalidClientRequestException, ServerException

        status_code = response.status_code
        detail = response.text
        code = None

        if detail.startswith('{'):
            data = json.loads(detail)
            detail = data['detail']
            code = data['code']

        if status_code == 404:
            raise ServiceNotFoundException(detail, code)

        if status_code in [403, 401]:
            raise PermissionDeniedException(detail, code)

        if status_code == 400:
            raise InvalidClientRequestException(detail, code)

        raise ServerException(detail, code)

    def build_url(self, endpoint: str) -> str:
        """
        Build url to use in the request

        Args:
            endpoint (str): the endpoint to use in the request

        Returns:
            str: the full url to use in the request
        """
        return f'{mesomb.host}/api/{mesomb.api_version}/{endpoint}'

    def get_authorization(self, method: str, endpoint: str, date: datetime, nonce: str,
                          headers: Optional[Dict[str, Any]] = None, body: Optional[Dict[str, Any]] = None) -> str:
        """
        Get the authorization to use in the request

        Args:
            method (str): the HTTP method to use
            endpoint (str): the endpoint to use in the request
            date (datetime): the date of the request
            nonce (str): the nonce of the request
            headers (Optional[Dict[str, Any]]): the headers to use in the request (Default value = None)
            body (Optional[Dict[str, Any]]): the body to use in the request (Default value = None)

        Returns:
            str: the authorization to use in the request
        """
        if headers is None:
            headers = {}

        url = self.build_url(endpoint)

        credentials = {'access_key': self.access_key, 'secret_key': self.secret_key}

        return Signature.sign_request(self.service, method, url, date, nonce, credentials, headers, body)

    def execute_request(self, method: str, endpoint: str, date: datetime, nonce: str = '', body: Dict[str, Any] = None,
                        mode: Optional[str] = None):
        """
        Execute the request to the MeSomb API

        Args:
            method (str): the HTTP method to use
            endpoint (str): the endpoint to use in the request
            date (datetime): the date of the request
            nonce (str): the nonce of the request (Default value = '')
            body (Dict[str, Any]): the body to use in the request (Default value = None)
            mode (Optional[str]): the mode to use in the request (Default value = None)

        Returns:
            dict: the response of the request

        Raises:
            ServiceNotFoundException: When the service is not found
            PermissionDeniedException: When the permission is denied
            InvalidClientRequestException: When the client request is invalid
            ServerException: When the server return an error
        """
        url = self.build_url(endpoint)

        headers = {
            'x-mesomb-date': str(int(date.timestamp())),
            'x-mesomb-nonce': nonce,
            'Accept-Language': self.language,
            'X-MeSomb-Source': f'PyMeSomb/{__version__}',
        }

        if self.service == 'payment':
            headers['X-MeSomb-Application'] = self.target

        if self.service == 'fundraising':
            headers['X-MeSomb-Fund'] = self.target

        if self.service == 'wallet':
            headers['X-MeSomb-Provider'] = self.target

        if body and 'trxID' in body:
            headers['X-MeSomb-TrxID'] = str(body.pop('trxID'))

        if mode:
            headers['X-MeSomb-OperationMode'] = mode

        if method == 'POST':
            authorization = self.get_authorization(method, endpoint, date, nonce,
                                                   headers={'content-type': 'application/json'},
                                                   body=body)
        else:
            authorization = self.get_authorization(method, endpoint, date, nonce)

        headers['Authorization'] = authorization

        response = requests.request(method, url, json=body, headers=headers)

        status_code = response.status_code
        if status_code >= 400:
            self.process_client_exception(response)

        return response.json()


class PaymentOperation(AOperation):
    """
    Payment operation class to interact with the MeSomb payment service
    """
    service = 'payment'

    def __init__(self, application_key, access_key, secret_key):
        super().__init__(application_key, access_key, secret_key)

    def make_collect(self, amount: float, service: str, payer: str, nonce: Optional[str] = None, country: str = 'CM',
                     currency: str = 'XAF', fees: bool = True, mode: str = 'synchronous', conversion: bool = False,
                     location: Optional[Dict[str, str]] = None, products: Optional[List[Dict[str, str]]] = None,
                     customer: Optional[Dict[str, str]] = None, trx_id: Optional[str] = None) -> TransactionResponse:
        """Collect money a use account

        Args:
            amount (float): amount to collect
            service (str): payment service with the possible values MTN, ORANGE, AIRTEL
            payer (str): account number to collect money
            nonce (str, optional): unique string on each request
            country (str): 2 letters country code of the service, configured during your service registration
                        in MeSomb (Default value = 'CM')
            currency (str): currency of your service depending on your country (Default value = 'XAF')
            fees (bool): false if your want MeSomb fees to be computed and included in the amount to collect
                        (Default value = True)
            conversion (bool): true in case of foreign currently defined if you want to rely on MeSomb to convert
                        the amount in the local currency (Default value = False)
            location (dict): Map containing the location of the customer with the following attributes:
                        town, region and location all string.
            products (list): It is array of products. Each product are Map with the following attributes:
                        name string, category string, quantity int and amount float
            customer (dict): a Map containing information about the customer: phone string, email: string, first_name
                        string, last_name string, address string, town string, region string and country string
            trx_id (str): if you want to include your transaction ID in the request
            mode (str): Mode of the transaction synchronous or asynchronous (Default value = 'synchronous')

        Returns:
            TransactionResponse
        """

        endpoint = 'payment/collect/'

        body = {
            'amount': amount,
            'payer': payer,
            'fees': fees,
            'service': service,
            'country': country,
            'currency': currency,
            'amount_currency': currency,
            'conversion': conversion
        }
        if trx_id:
            body['trxID'] = trx_id

        if location:
            body['location'] = location

        if customer:
            body['customer'] = customer

        if products:
            body['products'] = products

        return TransactionResponse(
            self.execute_request('POST', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body, mode))

    def make_deposit(self, amount: float, service: str, receiver: str, nonce: Optional[str] = None,
                     country: Optional[str] = 'CM', currency: Optional[str] = 'XAF', conversion: Optional[bool] = False,
                     location: Optional[Dict[str, str]] = None, products: Optional[List[Dict[str, str]]] = None,
                     customer: Optional[Dict[str, str]] = None, trx_id: Optional[str] = None) -> TransactionResponse:
        """
        Method to make deposit in a receiver mobile account.

        Args:
            amount (float): amount to collect
            service (str): payment service with the possible values MTN, ORANGE, AIRTEL
            receiver (str): account number to depose money
            nonce (str, optional): unique string on each request
            country (str, optional): 2 letters country code of the service configured during your service
                    registration in MeSomb (Default value = 'CM')
            currency (str, optional): currency of your service depending on your country (Default value = 'XAF')
            conversion (bool, optional): true in case of foreign currently defined if you want to rely on
                    MeSomb to convert the amount in the local currency (Default value = False)
            location (dict): Map containing the location of the customer with the following attributes:
                    town, region and location all string.
            products (list): It is array of products. Each product are Map with the following attributes:
                    name string, category string, quantity int and amount float
            customer (dict): a Map containing information about the customer: phone string, email: string,
                    first_name string, last_name string, address string, town string, region string and country string
            trx_id (str): if you want to include your transaction ID in the request

        Returns:
            TransactionResponse
        """
        endpoint = 'payment/deposit/'

        body = {
            'amount': amount,
            'receiver': receiver,
            'service': service,
            'country': country,
            'currency': currency,
            'amount_currency': currency,
            'conversion': conversion,
        }

        if trx_id:
            body['trxID'] = trx_id

        if location:
            body['location'] = location

        if customer:
            body['customer'] = customer

        if products:
            body['products'] = products

        return TransactionResponse(
            self.execute_request('POST', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body))

    def make_yango_refill(self, amount: float, service: str, payer: str, driver_id: str, nonce=None,
                          country: Optional[str] = 'CM', currency: Optional[str] = 'XAF',
                          mode: Optional[str] = 'synchronous', location: Optional[Dict[str, str]] = None,
                          customer: Optional[Dict[str, str]] = None,
                          trx_id: Optional[str] = None) -> TransactionResponse:
        """
        Method to refill a Yango account.

        Args:
            amount (float): amount to collect
            service (str): payment service with the possible values MTN, ORANGE, AIRTEL
            payer (str): account number to depose money
            driver_id (str): the driver ID
            nonce (str, optional): unique string on each request
            country (str): 2 letters country code of the service configured during your service registration in
                    MeSomb (Default value = 'CM')
            currency (str): currency of your service depending on your country (Default value = 'XAF')
            mode (str): Mode of the transaction synchronous or asynchronous (Default value = 'synchronous')
            location (dict): Map containing the location of the customer with the following attributes:
                    town, region and location all string.
            customer (dict): a Map containing information about the customer: phone string, email: string,
                    first_name string, last_name string, address string, town string, region string and country string
            trx_id (str): if you want to include your transaction ID in the request

        Returns:
          TransactionResponse

        """
        endpoint = 'payment/yango/refill/'

        body = {
            'amount': amount,
            'payer': payer,
            'service': service,
            'driver_id': driver_id,
            'country': country,
            'currency': currency,
            'amount_currency': currency,
        }

        if trx_id:
            body['trxID'] = trx_id

        if location:
            body['location'] = location

        if customer:
            body['customer'] = customer

        return TransactionResponse(
            self.execute_request('POST', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body, mode))

    def get_status(self) -> Application:
        """Get the current status of your service on MeSomb

        Returns:
          Application

        """
        endpoint = 'payment/status/'

        return Application(self.execute_request('GET', endpoint, datetime.now()))

    def get_transactions(self, ids, source='MESOMB') -> List[Transaction]:
        """
        Get transactions from MeSomb by IDs.

        Args:
            source: source of the transactionID: MESOMB or EXTERNAL (Default value = 'MESOMB')
            ids: list of ids

        Returns:
            List[Transaction]
        """
        endpoint = f"payment/transactions/?{'&'.join([f'ids={id}' for id in ids])}&source={source}"

        return [Transaction(item) for item in self.execute_request('GET', endpoint, datetime.now())]

    def check_transactions(self, ids, source='MESOMB') -> List[Transaction]:
        """
        Check transactions from MeSomb by IDs.

        Args:
            source: source of the transactionID: MESOMB or EXTERNAL (Default value = 'MESOMB')
            ids: list of ids

        Returns:
            List[Transaction]
        """
        endpoint = f"payment/transactions/check/?ids={','.join(ids)}&source={source}"
        return [Transaction(item) for item in self.execute_request('GET', endpoint, datetime.now())]

    def refund_transaction(self, trx_id: str, amount: Optional[float] = None, conversion: Optional[bool] = None,
                           currency: str = None) -> TransactionResponse:
        """
        Refund a transaction in MeSomb

        Args:
            trx_id: the transaction identifier
            amount: the amount to refund
            conversion: true if you want to rely on MeSomb to convert the amount in the local
                currency (Default value = None)
            currency: currency of your service depending on your country (Default value = None)
        Returns:
            TransactionResponse
        """
        endpoint = f'payment/refund/'

        body = {'id': trx_id}
        if amount:
            body['amount'] = amount
        if currency:
            body['currency'] = currency
            body['amount_currency'] = currency
        if conversion:
            body['conversion'] = conversion

        return TransactionResponse(self.execute_request('POST', endpoint, datetime.now(), RandomGenerator.nonce(),
                                                       body))


class WalletOperation(AOperation):
    """
    Wallet operation class to interact with the MeSomb wallet service
    """
    service = 'wallet'

    def __init__(self, provider_key, access_key, secret_key):
        super().__init__(provider_key, access_key, secret_key)

    def create_wallet(self, last_name: str, phone_number: str, gender: str, first_name: Optional[str] = None,
                      country: Optional[str] = 'CM', email: Optional[str] = None, nonce: Optional[str] = None,
                      number: Optional[str] = None):
        """
        Create a wallet in MeSomb

        Args:
            last_name (str): the last name of the wallet owner
            phone_number (str): the phone number of the wallet owner
            gender (str): the gender of the wallet owner
            first_name (str, optional): the first name of the wallet owner
            country (str, optional): the country of the wallet owner (Default value = 'CM')
            email (str, optional): the email of the wallet owner
            nonce (str, optional): the nonce of the request (Default value = None)
            number (str, optional): the unique numeric wallet identifier, if not set we will generate one for you

        Returns:
            Wallet
        """
        endpoint = 'wallet/wallets/'

        body = {
            'last_name': last_name,
            'phone_number': phone_number,
            'gender': gender,
            'country': country,
        }

        if first_name:
            body['first_name'] = first_name

        if email:
            body['email'] = email

        if number:
            body['number'] = number

        return Wallet(self.execute_request('POST', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body))

    def update_wallet(self, identifier: int, last_name: str, phone_number: str, gender: str,
                      first_name: Optional[str] = None, country: Optional[str] = 'CM', email: Optional[str] = None,
                      nonce: Optional[str] = None):
        """
        Create a wallet in MeSomb

        Args:
            identifier: the identifier of the wallet
            last_name (str): the last name of the wallet owner
            phone_number (str): the phone number of the wallet owner
            gender (str): the gender of the wallet owner
            first_name (str, optional): the first name of the wallet owner
            country (str, optional): the country of the wallet owner (Default value = 'CM')
            email (str, optional): the email of the wallet owner
            nonce (str, optional): the nonce of the request (Default value = None)

        Returns:
            Wallet
        """

        endpoint = f'wallet/wallets/{identifier}/'

        body = {
            'last_name': last_name,
            'phone_number': phone_number,
            'gender': gender,
            'country': country,
        }

        if first_name:
            body['first_name'] = first_name

        if email:
            body['email'] = email

        return Wallet(self.execute_request('PUT', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body))

    def get_wallet(self, identifier: int):
        """
        Get a wallet in MeSomb

        Args:
            identifier (int): the identifier of the wallet

        Returns:
            Wallet
        """
        endpoint = f'wallet/wallets/{identifier}/'

        return Wallet(self.execute_request('GET', endpoint, datetime.now()))

    def delete_wallet(self, identifier: int):
        """
        Delete a wallet in MeSomb

        Args:
            identifier (int): the identifier of the wallet

        Returns:
            None
        """
        endpoint = f'wallet/wallets/{identifier}/'

        return self.execute_request('DELETE', endpoint, datetime.now())

    def add_money(self, wallet: int, amount: float, message: Optional[str] = None, external_id: Optional[str] = None) -> WalletTransaction:
        """
        Add money to a wallet

        Args:
            wallet: the identifier of the wallet
            amount: the amount to add to the wallet
            message: the message to add to the transaction (Default value = None)

        Returns:
            Wallet
        """

        endpoint = f'wallet/wallets/{wallet}/adjust/'

        data = {'amount': amount, 'direction': 1}

        if message:
            data['message'] = message

        if external_id:
            data['trxID'] = external_id

        return WalletTransaction(self.execute_request('POST', endpoint, datetime.now(), RandomGenerator.nonce(), data))

    def remove_money(self, wallet: int, amount: float, force: Optional[bool] = False, message: Optional[str] = None, external_id: Optional[str] = None) -> WalletTransaction:
        """
        Remove money from a wallet

        Args:
            wallet: the identifier of the wallet
            amount: the amount to remove from the wallet
            force: to force the operation if balance is not enough (Default value = False)
            message: the message to add to the transaction (Default value = None)

        Returns:
            Wallet
        """
        endpoint = f'wallet/wallets/{wallet}/adjust/'

        data = {'amount': amount, 'direction': -1, 'force': force}

        if message:
            data['message'] = message

        if external_id:
            data['trxID'] = external_id

        return WalletTransaction(self.execute_request('POST', endpoint, datetime.now(), RandomGenerator.nonce(), data))

    def transfert_money(self, source: int, dest: int, amount: float,
                        force: Optional[bool] = False, message: Optional[str] = None, external_id: Optional[str] = None) -> WalletTransaction:
        """
        Transfer money from a wallet to another

        Args:
            source: the identifier of the source wallet
            dest: the identifier of the destination wallet
            amount: the amount to transfer
            force: to force the operation if balance is not enough (Default value = False)

        Returns:
            Wallet
        """
        endpoint = f'wallet/wallets/{source}/transfer/'

        data = {'amount': amount, 'destination': dest, 'force': force}

        if message:
            data['message'] = message

        if external_id:
            data['trxID'] = external_id

        return WalletTransaction(self.execute_request('POST', endpoint, datetime.now(), RandomGenerator.nonce(), data))

    def get_wallets(self, page=1):
        """
        Get wallets paginated

        Args:
            page: the page number (Default value = 1)

        Returns:
            PaginatedWallets
        """
        endpoint = f'wallet//wallets/?page={page}'

        return PaginatedWallets(self.execute_request('GET', endpoint, datetime.now()))

    def get_transaction(self, identifier: int):
        """
        Get a transaction in MeSomb

        Args:
            identifier (int): the identifier of the transaction

        Returns:
            Transaction
        """
        endpoint = f'wallet/transactions/{identifier}/'

        return WalletTransaction(self.execute_request('GET', endpoint, datetime.now()))

    def list_transactions(self, page: int = 1, wallet: Optional[int] = None):
        """
        Listing transactions from MeSomb

        Args:
            page (int): the page number (Default value = 1)
            wallet (int, optional): the identifier of the wallet (Default value = None)

        Returns:
            List[WalletTransaction]
        """
        endpoint = f'wallet/transactions/?page={page}'

        if wallet:
            endpoint += f'&wallet={wallet}'

        return PaginatedWalletTransactions(self.execute_request('GET', endpoint, datetime.now()))

    def get_transactions(self, ids, source='MESOMB') -> List[WalletTransaction]:
        """
        Get transactions base on external in MeSomb's IDs

        Args:
            ids: list of ids
            source: source of transactions ids MESOMB or EXTERNAL

        Returns:
            List[WalletTransaction]
        """
        endpoint = f"wallet/transactions/search/?{'&'.join([f'ids={id}' for id in ids])}&source={source}"

        return [WalletTransaction(t) for t in self.execute_request('GET', endpoint, datetime.now())]


class FundraisingOperation(AOperation):
    """
    Fundraising operation class to interact with the MeSomb fundraising service
    """
    service = 'fundraising'

    def make_contribution(self, amount: float, service: str, payer: str, nonce: Optional[str] = None,
                          country: str = 'CM',currency: str = 'XAF', mode: str = 'synchronous',
                          conversion: bool = False, anonymous: bool = False, accept_terms: bool = True,
                          location: Optional[Dict[str, str]] = None, contact: Optional[Dict[str, str]] = None,
                          full_name: Optional[Dict[str, str]] = None,
                          trx_id: Optional[str] = None) -> ContributionResponse:
        """
        Make a contribution to a fundraising campaign

        Args:
            amount: amount to contribute
            service: payment service with the possible values MTN, ORANGE, AIRTEL
            payer: account number to contribute
            nonce: unique string on each request
            country: 2 letters country code of the service, configured during your service registration in MeSomb
            currency: currency of your service depending on your country
            conversion: true in case of foreign currently defined if you want to rely on MeSomb to
                convert the amount in the local currency
            anonymous: true if you want to contribute anonymously
            accept_terms: true if you accept the terms of the contribution
            location: Map containing the location of the customer with the following attributes:
                town, region and location all string.
            contact: a Map containing information about the customer: phone_number string, email: string
            full_name: a Map containing information about the customer: first_name string, last_name string
            trx_id: if you want to include your transaction ID in the request

        Returns:
            ContributionResponse

        """
        assert amount > 0, 'Amount must be greater than 0'
        if not anonymous:
            assert full_name, 'Full name is required'
            assert contact, 'Contact is required'

        endpoint = 'fundraising/contribute/'

        body = {
            'amount': amount,
            'payer': payer,
            'service': service,
            'country': country,
            'currency': currency,
            'amount_currency': currency,
            'conversion': conversion,
            'anonymous': anonymous,
            'accept_terms': accept_terms,
        }

        if trx_id:
            body['trxID'] = trx_id

        if location:
            body['location'] = location

        if contact:
            body['contact'] = contact

        if full_name:
            body['full_name'] = full_name

        return ContributionResponse(
            self.execute_request('POST', endpoint, datetime.now(), nonce or RandomGenerator.nonce(), body, mode))

    def get_contributions(self, ids, source='MESOMB') -> List[Contribution]:
        """
        Get contributions from MeSomb by IDs.

        Args:
            source: source of the contributionID: MESOMB or EXTERNAL (Default value = 'MESOMB')
            ids: list of ids

        Returns:
            List[Contribution]
        """
        assert source in ['MESOMB', 'EXTERNAL'], 'Source must be MESOMB or EXTERNAL'

        endpoint = f"fundraising/contributions/?ids={','.join(ids)}&source={source}"

        return [Contribution(item) for item in self.execute_request('GET', endpoint, datetime.now())]

    def check_contributions(self, ids, source='MESOMB') -> List[Contribution]:
        """
        Check contributions from MeSomb by IDs.

        Args:
            source: source of the contributionID: MESOMB or EXTERNAL (Default value = 'MESOMB')
            ids: list of ids

        Returns:
            List[Contribution]
        """

        assert source in ['MESOMB', 'EXTERNAL'], 'Source must be MESOMB or EXTERNAL'

        endpoint = f"fundraising/contributions/check/?ids={','.join(ids)}&source={source}"
        return [Contribution(item) for item in self.execute_request('GET', endpoint, datetime.now())]



import unittest

from pymesomb import mesomb
from pymesomb.operations import WalletOperation


class WalletTest(unittest.TestCase):
    def setUp(self):
        mesomb.host = 'http://127.0.0.1:8000'
        self.provider_key = 'a1dc7a7391c538788043'
        self.access_key = 'c6c40b76-8119-4e93-81bf-bfb55417b392'
        self.secret_key ='fe8c2445-810f-4caa-95c9-778d51580163'

    def test_create_wallet(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.create_wallet(
            first_name='John',
            last_name='Doe',
            email='contact@gmail.com',
            phone_number='+237677550000',
            country='CM',
            gender='MAN'
        )

        self.assertEqual(wallet.first_name, 'John')
        self.assertEqual(wallet.last_name, 'Doe')
        self.assertEqual(wallet.email, 'contact@gmail.com')
        self.assertEqual(wallet.phone_number, '+237677550000')
        self.assertEqual(wallet.country, 'CM')
        self.assertEqual(wallet.gender, 'MAN')
        self.assertIsNotNone(wallet.number)

    def test_create_wallet_with_min_data(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.create_wallet(last_name='Doe', phone_number='+237677550000', gender='MAN')

        self.assertEqual(wallet.last_name, 'Doe')
        self.assertEqual(wallet.phone_number, '+237677550000')
        self.assertEqual(wallet.country, 'CM')
        self.assertEqual(wallet.gender, 'MAN')
        self.assertIsNotNone(wallet.number)

    def test_get_wallet(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.get_wallet(228)

        self.assertEqual(wallet.id, 228)
        self.assertEqual(wallet.first_name, 'John')
        self.assertEqual(wallet.last_name, 'Doe')
        self.assertEqual(wallet.email, 'contact@gmail.com')
        self.assertEqual(wallet.phone_number, '+237677550000')
        self.assertEqual(wallet.country, 'CM')
        self.assertEqual(wallet.gender, 'MAN')
        self.assertIsNotNone(wallet.number)

    def test_update_wallet(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.update_wallet(228, first_name='Jane', last_name='Doe', gender='MALE', phone_number='+237677550000')
        self.assertEqual(wallet.first_name, 'Jane')
        self.assertEqual(wallet.last_name, 'Doe')
        self.assertEqual(wallet.gender, 'MALE')
        self.assertEqual(wallet.phone_number, '+237677550000')

    def test_add_money_to_wallet(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.get_wallet(228)
        transaction = operation.add_money(228, amount=1000)

        self.assertEqual(transaction.direction, 1)
        self.assertEqual(transaction.status, 'SUCCESS')
        self.assertEqual(transaction.amount, 1000)
        self.assertEqual(transaction.balance_after, (wallet.balance or 0) + 1000)
        self.assertEqual(transaction.wallet, 228)
        self.assertEqual(transaction.country, 'CM')
        self.assertIsNotNone(transaction.fin_trx_id)
        self.assertIsNotNone(transaction.date)

    def test_add_remove_to_wallet(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        wallet = operation.get_wallet(228)
        transaction = operation.remove_money(228, amount=1000)

        self.assertEqual(transaction.direction, -1)
        self.assertEqual(transaction.status, 'SUCCESS')
        self.assertEqual(transaction.amount, 1000)
        self.assertEqual(transaction.balance_after, (wallet.balance or 0) - 1000)
        self.assertEqual(transaction.wallet, 228)
        self.assertEqual(transaction.country, 'CM')
        self.assertIsNotNone(transaction.fin_trx_id)
        self.assertIsNotNone(transaction.date)

    def test_list_transactions(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        transactions = operation.list_transactions(1)

        self.assertGreater(transactions.count, 0)
        # self.assertIsNotNone(transactions.next)
        self.assertIsNone(transactions.previous)
        self.assertTrue(len(transactions.results) > 0)

    def test_get_transactions(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        transactions = operation.get_transactions([1047, 1048])

        self.assertGreater(len(transactions), 0)

        transactions = operation.get_transactions(['REF-1'], source='EXTERNAL')

        self.assertGreater(len(transactions), 0)

    def test_get_transaction(self):
        operation = WalletOperation(self.provider_key, self.access_key, self.secret_key)

        transaction = operation.get_transaction(3061)

        self.assertEqual(transaction.id, 3061)
        self.assertEqual(transaction.direction, -1)
        self.assertEqual(transaction.status, 'SUCCESS')
        self.assertEqual(transaction.amount, 1000)
        self.assertEqual(transaction.balance_after, 1000)
        self.assertEqual(transaction.wallet, 228)
        self.assertEqual(transaction.country, 'CM')
        self.assertIsNotNone(transaction.fin_trx_id)
        self.assertIsNotNone(transaction.date)


from abc import ABC
from datetime import datetime
from typing import Optional, List, Any, Dict


class Location:
    """
    Represents the location of a customer.

    Args:
        data (dict): The location data.

    Attributes:
        town (str): The town of the customer.
        region (str, optional): The region of the customer.
        country (str, optional): The country of the customer.
    """

    def __init__(self, data):
        self.town: str = data.get('town')
        self.region: Optional[str] = data.get('region')
        self.country: Optional[str] = data.get('country')


class Customer:
    """
    Represents a customer.

    Args:
        data (dict): The customer data.

    Attributes:
        email (str, optional): The email of the customer.
        phone (str, optional): The phone number of the customer.
        town (str, optional): The town of the customer.
        region (str, optional): The region of the customer.
        country (str, optional): The country of the customer.
        first_name (str, optional): The first name of the customer.
        last_name (str): The last name of the customer.
        address (str, optional): The address of the customer.
    """

    def __init__(self, data):
        self.email: Optional[str] = data.get('email', None)
        self.phone: Optional[str] = data.get('phone', None)
        self.town: Optional[str] = data.get('town', None)
        self.region: Optional[str] = data.get('region', None)
        self.country: Optional[str] = data.get('country', None)
        self.first_name: Optional[str] = data.get('first_name', None)
        self.last_name: str = data['last_name']
        self.address: Optional[str] = data.get('address', None)


class Product:
    """
    Represents a product.

    Args:
        data (dict): The product data.

    Attributes:
        id (str): The identifier of the product.
        name (str): The name of the product.
        category (str, optional): The category of the product.
        quantity (int, optional): The quantity of the product.
        amount (float, optional): The amount of the product.
    """

    def __init__(self, data):
        self.id: str = data['id']
        self.name: str = data['name']
        self.category: Optional[str] = data.get('category', 'category')
        self.quantity: Optional[int] = data.get('quantity', None)
        self.amount: Optional[float] = data.get('amount', None)


class ATransaction(ABC):
    """
    Represents a transaction.

    Args:
        data (dict): The transaction data.

    Attributes:
        pk (str): The primary key of the transaction.
        status (str): The status of the transaction.
        type (str): The type of the transaction.
        amount (float): The amount of the transaction.
        fees (float, optional): The fees of the transaction.
        b_party (str): The b party of the transaction.
        message (str, optional): The message of the transaction.
        service (str): The service of the transaction.
        reference (str, optional): The reference of the transaction.
        date (datetime): The timestamp of the transaction.
        country (str): The country of the transaction.
        currency (str): The currency of the transaction.
        fin_trx_id (str, optional): The financial transaction ID of the transaction.
        trxamount (float, optional): The transaction amount.
        location (Location, optional): The location of the transaction.
    """

    def __init__(self, data: Dict[str, Any]):
        self.pk: str = data['pk']
        self.status: str = data['status']
        self.type: str = data['type']
        self.amount: float = data['amount']
        self.fees: Optional[float] = data.get('fees')
        self.b_party: str = data['b_party']
        self.message: Optional[str] = data.get('message')
        self.service: str = data['service']
        self.reference: Optional[str] = data.get('reference')
        self.date: datetime = datetime.strptime(data['ts'], '%Y-%m-%dT%H:%M:%SZ')
        self.country: str = data['country']
        self.currency: str = data['currency']
        self.fin_trx_id: Optional[str] = data.get('fin_trx_id')
        self.trxamount: Optional[float] = data.get('trxamount')
        self.location: Optional[Location] = Location(data['location']) if data.get('location') else None

    def is_success(self) -> bool:
        """
        Check if the transaction was successful.

        Returns:
            bool: True if the transaction was successful, False otherwise.
        """
        return self.status == 'SUCCESS'

    def is_pending(self) -> bool:
        """
        Check if the transaction is pending.

        Returns:
            bool: True if the transaction is pending, False otherwise.
        """
        return self.status == 'PENDING'

    def is_failed(self) -> bool:
        """
        Check if the transaction failed.

        Returns:
            bool: True if the transaction failed, False otherwise.
        """
        return self.status == 'FAILED'


class Transaction(ATransaction):
    """
    Represents a transaction.
    This class extends :ATransaction:. It adds more attributes to the transaction.

    Args:
        data (dict): The transaction data.

    Attributes:
        pk (str): The primary key of the transaction.
        status (str): The status of the transaction.
        type (str): The type of the transaction.
        amount (float): The amount of the transaction.
        fees (float, optional): The fees of the transaction.
        b_party (str): The b party of the transaction.
        message (str, optional): The message of the transaction.
        service (str): The service of the transaction.
        reference (str, optional): The reference of the transaction.
        date (datetime): The timestamp of the transaction.
        country (str): The country of the transaction.
        currency (str): The currency of the transaction.
        fin_trx_id (str, optional): The financial transaction ID of the transaction.
        trxamount (float, optional): The transaction amount.
        location (Location, optional): The location of the transaction.
        customer (Customer, optional): The customer of the transaction.
        products (List[Product], optional): The products of the transaction.
    """

    def __init__(self, data: dict):
        super().__init__(data)
        self.customer: Optional[Customer] = Customer(data['customer']) if data.get('customer') else None
        self.products: Optional[List[Product]] = [Product(p) for p in data.get('products', [])]
        # self.products = data['products?.map((d: Record < string, any >) = > new Product(d))']

    def __str__(self):
        return self.pk


class TransactionResponse:
    """
    Represents the response of a transactional operation.

    Args:
        data (dict): The response data from the server.

    Attributes:
        success (bool): Whether the operation was successful.
        message (str, optional): The message returned by the operation.
        redirect (str, optional): The URL to redirect the user to.
        transaction (Transaction): The transaction object.
        reference (str, optional): The reference of the transaction.
        status (str): The status of the transaction.
    """

    def __init__(self, data: Dict[str, Any]):
        self._data = data

        self.success: bool = data['success']
        self.message: Optional[str] = data['message']
        self.redirect: Optional[str] = data['redirect']
        self.transaction: Transaction = Transaction(data['transaction'])
        self.reference: Optional[str] = data['reference']
        self.status: str = data['status']

    def __str__(self):
        return '{}: {}'.format(self.status, self.message)

    def is_operation_success(self):
        """ """
        return self.success

    def is_transaction_success(self):
        """ """
        return self.transaction.is_success()

    def get_data(self) -> Dict[str, Any]:
        return self._data


class ApplicationBalance:
    """ """

    def __init__(self, data):
        self.country = data.get("country")
        self.currency = data.get("currency")
        self.provider = data.get("provider")
        self.value = data.get("value")
        self.service_name = data.get("service_name")


class Application:
    """
    Represents an application.

    Args:
        data (dict): The application data.

    Attributes:
        key (str): The key of the application.
        logo (str, optional): The logo of the application.
        balances (List[ApplicationBalance]): The balances of the application.
        countries (List[str]): The countries supported by the application.
        description (str, optional): The description of the application.
        name (str): The name of the application.
        security (dict, optional): The security of the application.
        url (str, optional): The URL of the application
    """

    def __init__(self, data: Dict[str, Any]):
        self.key: str = data['key']
        self.logo: Optional[str] = data.get('logo')
        self.balances: List[ApplicationBalance] = [ApplicationBalance(b) for b in data['balances']]
        self.countries: List[str] = data['countries']
        self.description: Optional[str] = data.get('description')
        self.name: str = data['name']
        self.security: Optional[Dict[str, Any]] = data.get('security')
        self.url: Optional[str] = data.get('url')

    def get_balance(self, country=None, service=None):
        """

        Args:
          country:  (Default value = None)
          service:  (Default value = None)

        Returns:

        """
        balance = 0
        for bal in self.balances:
            if country and bal.country != country:
                continue
            if service and bal.provider != service:
                continue
            balance += bal.value
        return balance


class WalletTransaction:
    """
    Represents a wallet transaction.

    Args:
        data (dict): The transaction data.

    Attributes:
        id (int): The identifier of the transaction.
        status (str): The status of the transaction.
        type (str): The type of the transaction.
        amount (float): The amount of the transaction.
        direction (int): The direction of the transaction.
        wallet (int): The wallet identifier.
        balance_after (float, optional): The balance after the transaction.
        date (datetime): The date of the transaction.
        country (str): The country of the transaction.
        fin_trx_id (str): The financial transaction ID.
    """

    def __init__(self, data: Dict[str, Any]):
        self.id: int = data['id']
        self.status: str = data['status']
        self.type: str = data['type']
        self.amount: float = data['amount']
        self.direction: int = data['direction']
        self.wallet: int = data['wallet']
        self.balance_after: Optional[float] = data.get('balance_after')
        self.date: datetime = datetime.strptime(data['date'], '%Y-%m-%dT%H:%M:%SZ')
        self.country: str = data['country']
        self.fin_trx_id: str = data['fin_trx_id']


class APaginated:
    """
    Represents a paginated response.

    Args:
        data (dict): The paginated data.

    Attributes:
        count (int): The total number of items.
        next (str, optional): The URL to the next page.
        previous (str, optional): The URL to the previous page.
    """

    def __init__(self, data: Dict[str, Any]):
        self.count: int = data['count']
        self.next: Optional[str] = data.get('next')
        self.previous: Optional[str] = data.get('previous')


class Wallet:
    """
    Represents a wallet.

    Args:
        data (dict): The wallet data.

    Attributes:
        id (int): The identifier of the wallet.
        number (str): The number of the wallet.
        country (str): The country of the wallet.
        status (str): The status of the wallet.
        last_activity (datetime, optional): The last activity of the wallet.
        balance (float): The balance of the wallet.
        first_name (str, optional): The first name of the wallet.
        last_name (str): The last name of the wallet.
        email (str, optional): The email of the wallet.
        phone_number (str): The phone number of the wallet.
    """

    def __init__(self, data: Dict[str, Any]):
        self.id: int = data['id']
        self.number: str = data['number']
        self.country: str = data['country']
        self.status: str = data['status']
        self.last_activity: Optional[datetime] = datetime.strptime(data['last_activity'], '%Y-%m-%dT%H:%M:%SZ') \
            if data.get('last_activity') else None
        self.balance: float = data['balance']
        self.first_name: Optional[str] = data['first_name']
        self.last_name: str = data['last_name']
        self.email: Optional[str] = data['email']
        self.phone_number: str = data['phone_number']
        self.gender = data['gender']

    def __str__(self):
        return self.number


class PaginatedWallets(APaginated):
    """
    Represents a paginated response of wallets.

    Args:
        data (dict): The paginated data.

    Attributes:
        results (List[Wallet]): The list of wallets.
    """

    def __init__(self, data: Dict[str, Any]):
        super().__init__(data)
        self.results: List[Wallet] = [Wallet(w) for w in data['results']]


class PaginatedWalletTransactions(APaginated):
    """
    Represents a paginated response of wallet transactions.

    Args:
        data (dict): The paginated data.

    Attributes:
        results (List[WalletTransaction]): The list of transactions.
    """

    def __init__(self, data: Dict[str, Any]):
        super().__init__(data)
        self.results: List[WalletTransaction] = [WalletTransaction(t) for t in data['results']]


class Contribution(ATransaction):
    """
    Represents a transaction.
    This class extends :ATransaction:. It adds more attributes to the transaction.

    Args:
        data (dict): The transaction data.

    Attributes:
        pk (str): The primary key of the transaction.
        status (str): The status of the transaction.
        type (str): The type of the transaction.
        amount (float): The amount of the transaction.
        fees (float, optional): The fees of the transaction.
        b_party (str): The b party of the transaction.
        message (str, optional): The message of the transaction.
        service (str): The service of the transaction.
        reference (str, optional): The reference of the transaction.
        date (datetime): The timestamp of the transaction.
        country (str): The country of the transaction.
        currency (str): The currency of the transaction.
        fin_trx_id (str, optional): The financial transaction ID of the transaction.
        trxamount (float, optional): The transaction amount.
        location (Location, optional): The location of the transaction.
        contributor (Customer, optional): The contributor of the transaction.
    """

    def __init__(self, data: dict):
        super().__init__(data)
        self.contributor: Optional[Customer] = Customer(data['contributor']) if data.get('contributor') else None

    def __str__(self):
        return self.pk


class ContributionResponse:
    """
    Represents the response of a transactional operation.

    Args:
        data (dict): The response data from the server.

    Attributes:
        success (bool): Whether the operation was successful.
        message (str, optional): The message returned by the operation.
        contribution (Contribution): The transaction object.
        status (str): The status of the transaction.
    """

    def __init__(self, data: Dict[str, Any]):
        self._data = data

        self.success: bool = data['success']
        self.message: Optional[str] = data['message']
        self.contribution: Contribution = Contribution(data['contribution'])
        self.status: str = data['status']

    def __str__(self):
        return '{}: {}'.format(self.status, self.message)

    def is_operation_success(self):
        """ """
        return self.success

    def is_contribution_success(self):
        """ """
        return self.contribution.is_success()

    def get_data(self) -> Dict[str, Any]:
        return self._data



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup(
    name='pymesomb',
    version='2.0.2',
    description='Python client for MeSomb services.',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(include=['pymesomb', 'pymesomb.*']),
    author='Hachther LLC',
    author_email='contact@hachther.com',
    keywords=['MeSomb', 'MobileMoney', 'OrangeMoney', 'YangoRefill'],
    url='https://github.com/hachther/mesomb-python-client.git',
    download_url='https://pypi.org/project/pymesomb/',
    install_requires=['requests'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ]
)
